@@ -1,11 +1,11 @@
 /*
  * which v2.x -- print full path of executables
- * Copyright (C) 1999, 2003, 2007  Carlo Wood <carlo@gnu.org>
+ * Copyright (C) 1999, 2003, 2007, 2008  Carlo Wood <carlo@gnu.org>
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * This program is free software: you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation, either version 3 of the License, or
+ * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -13,12 +13,12 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02110-1301, USA.
+ * along with this program. If not, see <http://www.gnu.org/licenses/>.
  */
 
 #include "sys.h"
 #include <stdio.h>
+#include <ctype.h>
 #include "getopt.h"
 #include "tilde/tilde.h"
 #include "bash.h"
@@ -27,24 +27,32 @@ static const char *progname;
 
 static void print_usage(FILE *out)
 {
-  fprintf(out, "Usage: %s [options] [--] programname [...]\n", progname);
-  fprintf(out, "Options: --version, -[vV] Print version and exit successfully.\n");
-  fprintf(out, "         --help,          Print this help and exit successfully.\n");
-  fprintf(out, "         --skip-dot       Skip directories in PATH that start with a dot.\n");
-  fprintf(out, "         --skip-tilde     Skip directories in PATH that start with a tilde.\n");
-  fprintf(out, "         --show-dot       Don't expand a dot to current directory in output.\n");
-  fprintf(out, "         --show-tilde     Output a tilde for HOME directory for non-root.\n");
-  fprintf(out, "         --tty-only       Stop processing options on the right if not on tty.\n");
-  fprintf(out, "         --all, -a        Print all matches in PATH, not just the first\n");
-  fprintf(out, "         --read-alias, -i Read list of aliases from stdin.\n");
-  fprintf(out, "         --skip-alias     Ignore option --read-alias; don't read stdin.\n");
-  fprintf(out, "         --read-functions Read shell functions from stdin.\n");
-  fprintf(out, "         --skip-functions Ignore option --read-functions; don't read stdin.\n");
+  fprintf(out, "Usage: %s [options] [--] COMMAND [...]\n", progname);
+  fprintf(out, "Write the full path of COMMAND(s) to standard output.\n\n");
+  fprintf(out, "  --version, -[vV] Print version and exit successfully.\n");
+  fprintf(out, "  --help,          Print this help and exit successfully.\n");
+  fprintf(out, "  --skip-dot       Skip directories in PATH that start with a dot.\n");
+  fprintf(out, "  --skip-tilde     Skip directories in PATH that start with a tilde.\n");
+  fprintf(out, "  --show-dot       Don't expand a dot to current directory in output.\n");
+  fprintf(out, "  --show-tilde     Output a tilde for HOME directory for non-root.\n");
+  fprintf(out, "  --tty-only       Stop processing options on the right if not on tty.\n");
+  fprintf(out, "  --all, -a        Print all matches in PATH, not just the first\n");
+  fprintf(out, "  --read-alias, -i Read list of aliases from stdin.\n");
+  fprintf(out, "  --skip-alias     Ignore option --read-alias; don't read stdin.\n");
+  fprintf(out, "  --read-functions Read shell functions from stdin.\n");
+  fprintf(out, "  --skip-functions Ignore option --read-functions; don't read stdin.\n\n");
+  fprintf(out, "Recommended use is to write the output of (alias; declare -f) to standard\n");
+  fprintf(out, "input, so that which can show aliases and shell functions. See which(1) for\n");
+  fprintf(out, "examples.\n\n");
+  fprintf(out, "If the options --read-alias and/or --read-functions are specified then the\n");
+  fprintf(out, "output can be a full alias or function definition, optionally followed by\n");
+  fprintf(out, "the full path of each command used inside of those.\n\n");
+  fprintf(out, "Report bugs to <which-bugs@gnu.org>.\n");
 }
 
 static void print_version(void)
 {
-  fprintf(stdout, "GNU which v" VERSION ", Copyright (C) 1999 - 2007 Carlo Wood.\n");
+  fprintf(stdout, "GNU which v" VERSION ", Copyright (C) 1999 - 2008 Carlo Wood.\n");
   fprintf(stdout, "GNU which comes with ABSOLUTELY NO WARRANTY;\n");
   fprintf(stdout, "This program is free software; your freedom to use, change\n");
   fprintf(stdout, "and distribute this program is protected by the GPL.\n");
@@ -203,7 +211,11 @@ static char *path_clean_up(const char *path)
 
   do
   {
-    if (!saw_slash || *p1 != '/')
+    /*
+     * Two leading slashes are allowed, having an OS implementation-defined meaning.
+     * See http://www.opengroup.org/onlinepubs/009695399/basedefs/xbd_chap04.html#tag_04_11
+     */
+    if (!saw_slash || *p1 != '/' || (p1 == path + 1 && p1[1] != '/'))
       *p2++ = *p1;
     if (saw_slash_dot && (*p1 == '/'))
       p2 -= 2;
@@ -498,25 +510,15 @@ int main(int argc, char *argv[])
     const char *h;
 
     if (!(h = getenv("HOME")))
+      h = sh_get_home_dir();
+
+    strncpy(home, h, sizeof(home));
+    home[sizeof(home) - 1] = 0;
+    homelen = strlen(home);
+    if (home[homelen - 1] != '/' && homelen < sizeof(home) - 1)
     {
-      fprintf(stderr, "%s: ", progname);
-      if (show_tilde)
-	fprintf(stderr, "--show-tilde");
-      else
-	fprintf(stderr, "--skip-tilde");
-      fprintf(stderr, ": Environment variable HOME not set\n");
-      show_tilde = skip_tilde = 0;
-    }
-    else
-    {
-      strncpy(home, h, sizeof(home));
-      home[sizeof(home) - 1] = 0;
-      homelen = strlen(home);
-      if (home[homelen - 1] != '/' && homelen < sizeof(home) - 1)
-      {
-	strcat(home, "/");
-	++homelen;
-      }
+      strcat(home, "/");
+      ++homelen;
     }
   }
 
